@@ -1,10 +1,15 @@
+from __future__ import annotations
+
 import json
 import secrets
 from collections import Counter
+from collections.abc import Iterable, Mapping, Sequence
 from datetime import UTC, datetime
 from functools import cached_property
+from typing import TYPE_CHECKING, Any, cast
 
 import sqlparse
+from opentelemetry.sdk.trace import ReadableSpan
 from opentelemetry.semconv._incubating.attributes import (
     exception_attributes,
     session_attributes,
@@ -52,48 +57,55 @@ class Trace(models.Model):
     share_id = models.CharField(max_length=32, default="", required=False)
     share_created_at = models.DateTimeField(allow_null=True, required=False)
 
-    class Meta:
-        ordering = ["-start_time"]
-        constraints = [
+    if TYPE_CHECKING:
+        from plain.models.fields.related_managers import BaseRelatedManager
+
+        spans: BaseRelatedManager
+        logs: BaseRelatedManager
+
+    model_options = models.Options(
+        ordering=["-start_time"],
+        constraints=[
             models.UniqueConstraint(
                 fields=["trace_id"],
                 name="observer_unique_trace_id",
             )
-        ]
-        indexes = [
+        ],
+        indexes=[
             models.Index(fields=["trace_id"]),
             models.Index(fields=["start_time"]),
             models.Index(fields=["request_id"]),
             models.Index(fields=["share_id"]),
             models.Index(fields=["session_id"]),
-        ]
+        ],
+    )
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self.trace_id
 
-    def get_absolute_url(self):
+    def get_absolute_url(self) -> str:
         """Return the canonical URL for this trace."""
         return reverse("observer:trace_detail", trace_id=self.trace_id)
 
-    def generate_share_id(self):
+    def generate_share_id(self) -> str:
         """Generate a unique share ID for this trace."""
         self.share_id = secrets.token_urlsafe(24)
         self.share_created_at = timezone.now()
         self.save(update_fields=["share_id", "share_created_at"])
         return self.share_id
 
-    def remove_share_id(self):
+    def remove_share_id(self) -> None:
         """Remove the share ID from this trace."""
         self.share_id = ""
         self.share_created_at = None
         self.save(update_fields=["share_id", "share_created_at"])
 
-    def duration_ms(self):
+    def duration_ms(self) -> float:
         return (self.end_time - self.start_time).total_seconds() * 1000
 
-    def get_trace_summary(self, spans):
+    def get_trace_summary(self, spans: Iterable[Span]) -> str:
         # Count database queries with query text and track duplicates
-        query_texts = []
+        query_texts: list[str] = []
         for span in spans:
             if query_text := span.attributes.get(db_attributes.DB_QUERY_TEXT):
                 query_texts.append(query_text)
@@ -103,7 +115,7 @@ class Trace(models.Model):
         duplicate_count = sum(query_counts.values()) - len(query_counts)
 
         # Build summary: "n spans, n queries (n duplicates), Xms"
-        parts = []
+        parts: list[str] = []
 
         # Queries count with duplicates
         if query_total > 0:
@@ -119,7 +131,7 @@ class Trace(models.Model):
         return " • ".join(parts)
 
     @classmethod
-    def from_opentelemetry_spans(cls, spans):
+    def from_opentelemetry_spans(cls, spans: Sequence[ReadableSpan]) -> Trace:
         """Create a Trace instance from a list of OpenTelemetry spans."""
         # Get trace information from the first span
         first_span = spans[0]
@@ -182,13 +194,15 @@ class Trace(models.Model):
             request_id=request_id,
             user_id=user_id,
             session_id=session_id,
-            app_name=app_name or getattr(settings, "APP_NAME", ""),
-            app_version=app_version or getattr(settings, "APP_VERSION", ""),
+            app_name=app_name or settings.NAME,
+            app_version=app_version or settings.VERSION,
             root_span_name=root_span.name if root_span else "",
         )
 
-    def as_dict(self):
-        spans = [span.span_data for span in self.spans.all().order_by("start_time")]
+    def as_dict(self) -> dict[str, Any]:
+        spans = [
+            span.span_data for span in self.spans.query.all().order_by("start_time")
+        ]
         logs = [
             {
                 "timestamp": log.timestamp.isoformat(),
@@ -196,7 +210,7 @@ class Trace(models.Model):
                 "message": log.message,
                 "span_id": log.span_id,
             }
-            for log in self.logs.all().order_by("timestamp")
+            for log in self.logs.query.all().order_by("timestamp")
         ]
 
         return {
@@ -215,11 +229,11 @@ class Trace(models.Model):
             "logs": logs,
         }
 
-    def get_timeline_events(self):
+    def get_timeline_events(self) -> list[dict[str, Any]]:
         """Get chronological list of spans and logs for unified timeline display."""
-        events = []
+        events: list[dict[str, Any]] = []
 
-        for span in self.spans.all().annotate_spans():
+        for span in self.spans.query.all().annotate_spans():
             events.append(
                 {
                     "type": "span",
@@ -230,7 +244,7 @@ class Trace(models.Model):
             )
 
             # Add logs for this span
-            for log in self.logs.filter(span=span):
+            for log in self.logs.query.filter(span=span):
                 events.append(
                     {
                         "type": "log",
@@ -241,7 +255,7 @@ class Trace(models.Model):
                 )
 
         # Add unlinked logs (logs without span)
-        for log in self.logs.filter(span__isnull=True):
+        for log in self.logs.query.filter(span__isnull=True):
             events.append(
                 {
                     "type": "log",
@@ -255,13 +269,13 @@ class Trace(models.Model):
         return sorted(events, key=lambda x: x["timestamp"])
 
 
-class SpanQuerySet(models.QuerySet):
-    def annotate_spans(self):
+class SpanQuerySet(models.QuerySet["Span"]):
+    def annotate_spans(self) -> list[Span]:
         """Annotate spans with nesting levels and duplicate query warnings."""
-        spans = list(self.order_by("start_time"))
+        spans: list[Span] = list(self.order_by("start_time"))
 
         # Build span dictionary for parent lookups
-        span_dict = {span.span_id: span for span in spans}
+        span_dict: dict[str, Span] = {span.span_id: span for span in spans}
 
         # Calculate nesting levels
         for span in spans:
@@ -273,7 +287,7 @@ class SpanQuerySet(models.QuerySet):
                 parent_level = parent.level if parent else 0
                 span.level = parent_level + 1
 
-        query_counts = {}
+        query_counts: dict[str, int] = {}
 
         # First pass: count queries
         for span in spans:
@@ -281,7 +295,7 @@ class SpanQuerySet(models.QuerySet):
                 query_counts[sql_query] = query_counts.get(sql_query, 0) + 1
 
         # Second pass: add annotations
-        query_occurrences = {}
+        query_occurrences: dict[str, int] = {}
         for span in spans:
             span.annotations = []
 
@@ -316,25 +330,30 @@ class Span(models.Model):
     status = models.CharField(max_length=50, default="", required=False)
     span_data = models.JSONField(default=dict, required=False)
 
-    objects = SpanQuerySet.as_manager()
+    query = SpanQuerySet()
 
-    class Meta:
-        ordering = ["-start_time"]
-        constraints = [
+    model_options = models.Options(
+        ordering=["-start_time"],
+        constraints=[
             models.UniqueConstraint(
                 fields=["trace", "span_id"],
                 name="observer_unique_span_id",
             )
-        ]
-        indexes = [
+        ],
+        indexes=[
             models.Index(fields=["span_id"]),
             models.Index(fields=["trace", "span_id"]),
             models.Index(fields=["trace"]),
             models.Index(fields=["start_time"]),
-        ]
+        ],
+    )
+
+    if TYPE_CHECKING:
+        level: int
+        annotations: list[dict[str, Any]]
 
     @classmethod
-    def from_opentelemetry_span(cls, otel_span, trace):
+    def from_opentelemetry_span(cls, otel_span: ReadableSpan, trace: Trace) -> Span:
         """Create a Span instance from an OpenTelemetry span."""
 
         span_data = json.loads(otel_span.to_json())
@@ -356,51 +375,51 @@ class Span(models.Model):
             span_data=span_data,
         )
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self.span_id
 
     @property
-    def attributes(self):
+    def attributes(self) -> Mapping[str, Any]:
         """Get attributes from span_data."""
-        return self.span_data.get("attributes", {})
+        return cast(Mapping[str, Any], self.span_data.get("attributes", {}))
 
     @property
-    def events(self):
+    def events(self) -> list[Mapping[str, Any]]:
         """Get events from span_data."""
-        return self.span_data.get("events", [])
+        return cast(list[Mapping[str, Any]], self.span_data.get("events", []))
 
     @property
-    def links(self):
+    def links(self) -> list[Mapping[str, Any]]:
         """Get links from span_data."""
-        return self.span_data.get("links", [])
+        return cast(list[Mapping[str, Any]], self.span_data.get("links", []))
 
     @property
-    def resource(self):
+    def resource(self) -> Mapping[str, Any]:
         """Get resource from span_data."""
-        return self.span_data.get("resource", {})
+        return cast(Mapping[str, Any], self.span_data.get("resource", {}))
 
     @property
-    def context(self):
+    def context(self) -> Mapping[str, Any]:
         """Get context from span_data."""
-        return self.span_data.get("context", {})
+        return cast(Mapping[str, Any], self.span_data.get("context", {}))
 
-    def duration_ms(self):
+    def duration_ms(self) -> float:
         if self.start_time and self.end_time:
             return (self.end_time - self.start_time).total_seconds() * 1000
         return 0
 
     @cached_property
-    def sql_query(self):
+    def sql_query(self) -> str | None:
         """Get the SQL query if this span contains one."""
         return self.attributes.get(db_attributes.DB_QUERY_TEXT)
 
     @cached_property
-    def sql_query_params(self):
+    def sql_query_params(self) -> dict[str, Any]:
         """Get query parameters from attributes that start with 'db.query.parameter.'"""
         if not self.attributes:
             return {}
 
-        query_params = {}
+        query_params: dict[str, Any] = {}
         for key, value in self.attributes.items():
             if key.startswith(DB_QUERY_PARAMETER_TEMPLATE + "."):
                 param_name = key.replace(DB_QUERY_PARAMETER_TEMPLATE + ".", "")
@@ -409,7 +428,7 @@ class Span(models.Model):
         return query_params
 
     @cached_property
-    def source_code_location(self):
+    def source_code_location(self) -> dict[str, Any] | None:
         """Get the source code location attributes from this span."""
         if not self.attributes:
             return None
@@ -431,7 +450,7 @@ class Span(models.Model):
 
         return code_attrs if code_attrs else None
 
-    def get_formatted_sql(self):
+    def get_formatted_sql(self) -> str | None:
         """Get the pretty-formatted SQL query if this span contains one."""
         sql = self.sql_query
         if not sql:
@@ -449,22 +468,27 @@ class Span(models.Model):
             comma_first=False,
         )
 
-    def format_event_timestamp(self, timestamp):
+    def format_event_timestamp(
+        self, timestamp: float | int | datetime | str
+    ) -> datetime | str:
         """Convert event timestamp to a readable datetime."""
         if isinstance(timestamp, int | float):
+            ts_value = float(timestamp)
             try:
                 # Try as seconds first
-                if timestamp > 1e10:  # Likely nanoseconds
-                    timestamp = timestamp / 1e9
-                elif timestamp > 1e7:  # Likely milliseconds
-                    timestamp = timestamp / 1e3
+                if ts_value > 1e10:  # Likely nanoseconds
+                    ts_value /= 1e9
+                elif ts_value > 1e7:  # Likely milliseconds
+                    ts_value /= 1e3
 
-                return datetime.fromtimestamp(timestamp, tz=UTC)
+                return datetime.fromtimestamp(ts_value, tz=UTC)
             except (ValueError, OSError):
-                return str(timestamp)
-        return timestamp
+                return str(ts_value)
+        if isinstance(timestamp, datetime):
+            return timestamp
+        return str(timestamp)
 
-    def get_exception_stacktrace(self):
+    def get_exception_stacktrace(self) -> str | None:
         """Get the exception stacktrace if this span has an exception event."""
         if not self.events:
             return None
@@ -492,11 +516,12 @@ class Log(models.Model):
     level = models.CharField(max_length=20)
     message = models.TextField()
 
-    class Meta:
-        ordering = ["timestamp"]
-        indexes = [
+    model_options = models.Options(
+        ordering=["timestamp"],
+        indexes=[
             models.Index(fields=["trace", "timestamp"]),
             models.Index(fields=["trace", "span"]),
             models.Index(fields=["timestamp"]),
             models.Index(fields=["trace"]),
-        ]
+        ],
+    )
